/* Name of package */
#define PACKAGE "freesasa"

/* Define to the full name of this package. */
#define PACKAGE_NAME "FreeSASA"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "FreeSASA 2.0.2"

/* Define to the version of this package. */
#define PACKAGE_VERSION "2.0.2"

#define USE_XML 0

#define USE_JSON 0

#define USE_CHECK 0
